;
;	3D Julia Morpher
;	Written by Chris Dragan
;
;	This program should be compiled using Netwide Assembler this way:
;		nasm -o 3djulia.com 3djulia.asm
;
;	Nasm home page is "http://nasm.sourceforge.net/"
;	Chris Dragan's home page is "http://chris.dragan.name/"
;

%define MAXITERS	7		; Limit of iterations
%define PRECISION	6		; Precision
%define HIGHESTBAR	30		; Height of highest bar
%define NUMCOLORS	10		; Number of used colors
%define SIDECOLOR	2		; Color index of bar's side

org 100h

section .text

;
;	Checking MMX support presence
;
	mov	ax, 7202h			; check if 386+
	push	ax
	popf
	pushf
	pop	ax
	cmp	ah, 0F0h			; XT
	jae	_nommx
	test	ah, 0F0h			; 80286
	jnz	_386

_nommx: mov	ah, 09h 			; output error message
	mov	dx, szNoMMX
	int	21h
	int	20h				; quit

_386:	pushfd					; check CPUID presence
	pop	eax
	mov	ecx, eax
	btc	eax, 21
	push	eax
	popfd
	pushfd
	pop	eax
	xor	eax, ecx
	bt	eax, 21
	jnc	_nommx

	xor	eax, eax			; check command 1 presence
	cpuid
	or	ax, ax
	jz	_nommx

	mov	eax, 1				; check mmx presence
	cpuid
	bt	edx, 23
	jnc	_nommx

;
;	Setting video mode 13h (320x200x256) and palette
;

	mov	ax, 0013h			; change video mode
	int	10h

	mov	dx, 03C8h			; setup palette
	mov	al, 01h
	out	dx, al
	inc	dl
	mov	cx, NUMCOLORS*3
	mov	si, Colors
	cld
	rep	outsb

	push	word 0A000h			; clear constant part of screen
	pop	es
	xor	di, di
	mov	cx, (320*(100-HIGHESTBAR))>> 2
	mov	eax, 01010101h
	rep	stosd
	push	ds
	pop	es

;
;	Loop start - checking keyboard buffer
;	and exiting when a key was pressed
;

lb_01:
;	hlt					; wait for PIT tick - synchro
	mov	ah, 01h 			; check keyboard buffer
	int	16h
	jz	lb_02

	mov	ah, 00h 			; read pressed key
	int	16h
	mov	ax, 0003h			; return to text mode
	int	10h
	emms					; empty mmx state
	int	20h				; quit

;
;	Heart of calculations
;

lb_02:	mov	di, FractalBuffer		; prepare to calculations
	mov	[Z0+4], dword -160
	mov	[Z0], dword -100
	movq	mm6, [CRe_CIm]
	movq	mm7, [_FFFF_0000_0000_0000]

lb_03:	movq	mm0, [Z0]			; calculation for next pixel
	mov	cx, MAXITERS

		; Loop optimised for Pentium MMX, 12 cycles
lb_04:			packssdw mm0, mm0	; 1 U

			movq	mm1, mm0	; 2 U
			pxor	mm0, mm7	;   V

			psubw	mm0, mm7	; 3 U
			movq	mm2, mm1	;   V

			pmaddwd mm0, mm1	; 4 U
			psllq	mm1, 16 	;   V

			pmaddwd mm1, mm2	; 5 U
			nop			;   V

			;-			; 6 - last cycle of mul mm0

			;-			; 7 - mm0 being written

			movd	eax, mm0	; 8 U
			punpckhdq mm0, mm1	;   V

			cmp	eax, 4 << (2*PRECISION)
			jg	lb_05

			psrad	mm0, PRECISION	; 10
			paddd	mm0, mm6	; 11

			dec	cx		; 12 U
			jnz	lb_04		;    V

lb_05:	mov	al, MAXITERS			; store color for current pixel
	sub	al, cl
	stosb

	inc	dword [Z0+4]			; move to next pixel
	cmp	[Z0+4], dword 160
	jl	lb_03
	mov	[Z0+4], dword -160
	add	[Z0], dword 2
	test	[Z0+3], byte 80h
	jnz	lb_03

;
;	Drawing
;

	mov	di, ScreenBuffer		; clear drawing buffer
	mov	cx, (320*(100+HIGHESTBAR))>> 2
	mov	eax, 01010101h
	rep	stosd

	mov	si, FractalBuffer		; prepare to drawing loop
	mov	di, ScreenBuffer+(320*HIGHESTBAR)
	mov	cx, 320*50
	mov	ax, 1
	xor	bx, bx

lb_06:	mov	bl, [si]			; get bar index
	add	si, ax
	push	cx
	push	di
	mov	dl, [BarColors+bx]		; get bar top color
	shl	bx, 1
	mov	cx, [BarHeights+bx]		; get bar height
lb_07:	mov	[di], byte SIDECOLOR		; draw bar side
	sub	di, 320
	loop	lb_07
	mov	[di], dl			; draw bar top
	pop	di				; next pixel/bar
	pop	cx
	inc	di
	loop	lb_06

	cmp	ax, 1				; now draw inverted
	jnz	lb_08
	dec	si
	neg	ax
	mov	cx, 320*50
	jmp	lb_06

lb_08:	push	word 0A000h			; update screen
	pop	es
	mov	di, 320*(100-HIGHESTBAR)
	mov	cx, (320*(100+HIGHESTBAR))>> 2
	mov	si, ScreenBuffer
	rep	movsd
	push	ds
	pop	es

;
;	3D Julia animation
;

	mov	eax, [CRe]			; update C.re
	mov	ecx, [HorizSpeed]
	add	eax, ecx
	cmp	eax, 160
	jl	lb_09
	mov	eax, 159
	jmp	lb_0A
lb_09:	cmp	eax, -160
	jnl	lb_0B
	mov	eax, -160
lb_0A:	neg	ecx
	mov	[HorizSpeed], ecx
lb_0B:	mov	[CRe], eax

	mov	eax, [CIm]			; update C.im
	mov	ecx, [VertSpeed]
	add	eax, ecx
	cmp	eax, 100
	jl	lb_0C
	mov	eax, 99
	jmp	lb_0D
lb_0C:	cmp	eax, -100
	jnl	lb_0E
	mov	eax, -100
lb_0D:	neg	ecx
	mov	[VertSpeed], ecx
lb_0E:	mov	[CIm], eax
	jmp	lb_01

section .data

szNoMMX 	db 'This program requires MMX support', 13, 10, '$'

Colors		db 20, 30, 0	; background
		db 20, 20, 0	; side
		db 40, 40, 10	; level 1
		db 42, 42, 12	; level 2
		db 44, 44, 14	; level 3
		db 46, 46, 16	; level 4
		db 48, 48, 18	; level 5
		db 50, 50, 20	; level 6
		db 55, 55, 22	; half
		db 60, 60, 22	; top

BarHeights	dw 2, 4, 6, 8, 11, 14, 20, 30
BarColors	db 3, 4, 5, 6, 7, 8, 9, 10

Z0		dd 0, 0
CRe_CIm:
CRe		dd 0
CIm		dd 0
HorizSpeed	dd 3
VertSpeed	dd 2

_FFFF_0000_0000_0000 dd 0, 0FFFF0000h

section .bss

FractalBuffer	resb 320*50
ScreenBuffer	resb 320*(100+HIGHESTBAR)
